(defpackage :blaster-common
  (:use :common-lisp)
  (:export #:game-map
           #:width #:height #:data
           #:print-map
           #:in-map-p
           #:entity
           #:x #:y
           #:bomb
           #:owner #:power #:ticks #:explosion-step
           #:bomb-kill-zone
           #:player
           #:dx #:dy #:name #:direction
           #:bomb-count #:bomb-power #:dropped-bombs
           #:moving-p #:status
           #:human
           #:robot
           #:+cell-size+ #:+cell-half-size+ #:+player-step+))

(in-package :blaster-common)

(defmacro define-simple-class (name superclasses slots)
  `(defclass ,name ,superclasses
    ,(mapcar (lambda (slot)
               (let ((keyword (intern (symbol-name slot) :keyword)))
               `(,slot :initarg ,keyword :accessor ,slot)))
             slots)))

(define-simple-class game-map ()
  (width height data))

(defun print-map (map)
  (loop for y below (height map)
        do (loop for x below (width map)
                 as cell = (aref (data map) y x)
                 do (princ (ecase cell
                             ((nil) #\space)
                             (:wall #\#)
                             (:obstacle #\*))))
        do (terpri)))

(defun in-map-p (x y map)
  (and (<= 0 x (1- (width map)))
       (<= 0 y (1- (height map)))))

(define-simple-class entity ()
  (x y))

(define-simple-class bomb (entity)
  (owner power ticks explosion-step))

(defun bomb-kill-zone (bomb map)
  (let ((cells (list (cons (x bomb) (y bomb)))))
    (loop for (dx . dy) in '((1 . 0) (-1 . 0) (0 . 1) (0 . -1))
          do (loop for i from 1
                   repeat (power bomb)
                   as x = (+ (x bomb) (* i dx))
                   as y = (+ (y bomb) (* i dy))
                   while (and (in-map-p x y map)
                              (not (eq (aref (data map) y x) :wall)))
                   do (push (cons x y) cells)
                   when (eq (aref (data map) y x) 'obstacle)
                   return nil))
    cells))

(define-simple-class player (entity)
  (dx dy
   name
   direction
   bomb-count bomb-power dropped-bombs
   moving-p
   status))

(define-simple-class human (player)
  ())

(define-simple-class robot (player)
  ())

(defparameter +cell-size+ 1000)

(defparameter +cell-half-size+ (floor +cell-size+ 2))

(defparameter +player-step+ (floor +cell-size+ 5))
